@@ -1,12 +1,18 @@
 #include "canl_locl.h"
 #include "canl_cred.h"
 #include "canl_mech_ssl.h"
+#include "scutils.h"
 
 #define DEF_KEY_LEN 1024
 #define DEF_KEY_LEN_LONGER 2048
 #define LIFETIME_TRESHOLD 10*24*60*60 //10 days
 
 static STACK_OF(X509)* my_sk_X509_dup(glb_ctx *cc, STACK_OF(X509) *stack);
+extern int proxy_verify_cert_chain(X509 * ucert, STACK_OF(X509) * cert_chain, proxy_verify_desc * pvd);
+extern proxy_verify_desc *pvd_setup_initializers(char *cadir);
+extern void pvd_destroy_initializers(void *data);
+extern canl_error map_verify_result(unsigned long ssl_err,
+                const X509_STORE_CTX *store_ctx, SSL *ssl);
 
 static STACK_OF(X509)* my_sk_X509_dup(glb_ctx *cc, STACK_OF(X509) *stack)
 {
@@ -104,11 +110,15 @@ canl_ctx_set_cred(canl_ctx ctx, canl_cred cred)
 {
     glb_ctx *cc = (glb_ctx*) ctx;
     creds *crd = (creds*) cred;
-    int ret = 0;
-    mech_glb_ctx *m_ctx = (mech_glb_ctx*) canl_mech_ssl.glb_ctx;
+    mech_glb_ctx *m_ctx = (mech_glb_ctx *)cc->mech_ctx;
 
     if (!ctx)
         return EINVAL;
+
+    if (!m_ctx)
+        return set_error(cc, EINVAL, POSIX_ERROR, "SSL context not"
+                " initialized");
+
     if (!crd || !m_ctx->cert_key)
         return set_error(cc, EINVAL, POSIX_ERROR, "Cred. handler"
                 " not initialized" );
@@ -122,11 +132,8 @@ canl_ctx_set_cred(canl_ctx ctx, canl_cred cred)
         }
     }
 
-    if (crd->c_key) {
-        if ((ret = pkey_dup(&m_ctx->cert_key->key, crd->c_key))) {
-            return ret;
-        }
-    }
+    if (crd->c_key)
+        pkey_dup(&m_ctx->cert_key->key, crd->c_key);
 
     if (crd->c_cert)
         m_ctx->cert_key->cert = X509_dup(crd->c_cert);
@@ -135,11 +142,10 @@ canl_ctx_set_cred(canl_ctx ctx, canl_cred cred)
     return 0;
 }
 
-int pkey_dup(EVP_PKEY **to, EVP_PKEY *from)
+void pkey_dup(EVP_PKEY **to, EVP_PKEY *from)
 {
     CRYPTO_add(&from->references,1,CRYPTO_LOCK_EVP_PKEY);
     *to = from;
-    return 0;
 }
 
 canl_err_code CANL_CALLCONV
@@ -164,6 +170,48 @@ canl_cred_load_priv_key_file(canl_ctx ctx, canl_cred cred, const char *pkey_file
     return ret;
 }
 
+canl_err_code CANL_CALLCONV
+canl_cred_save_priv_key(canl_ctx ctx, canl_cred cred, EVP_PKEY **pkey)
+{
+    glb_ctx *cc = (glb_ctx*) ctx;
+    creds *crd = (creds*) cred;
+    int ret = 0;
+
+    if (!ctx)
+        return EINVAL;
+
+    if (!cred)
+        return set_error(cc, EINVAL, POSIX_ERROR, "Cred. handler"
+                " not initialized" );
+    if (!pkey)
+        return set_error(cc, EINVAL, POSIX_ERROR, "Invalid private key"
+                " parameter");
+    pkey_dup(pkey, crd->c_key);
+
+    return ret;
+}
+
+canl_err_code CANL_CALLCONV
+canl_cred_load_priv_key(canl_ctx ctx, canl_cred cred, EVP_PKEY *pkey)
+{
+    glb_ctx *cc = (glb_ctx*) ctx;
+    creds *crd = (creds*) cred;
+    int ret = 0;
+
+    if (!ctx)
+        return EINVAL;
+
+    if (!cred)
+        return set_error(cc, EINVAL, POSIX_ERROR, "Cred. handler"
+                " not initialized" );
+    if (!pkey)
+        return set_error(cc, EINVAL, POSIX_ERROR, "Invalid private key"
+                " parameter");
+    pkey_dup(&crd->c_key, pkey);
+
+    return ret;
+}
+
 canl_err_code CANL_CALLCONV
 canl_cred_load_chain(canl_ctx ctx, canl_cred cred, STACK_OF(X509) *cert_stack)
 {
@@ -355,7 +403,7 @@ canl_cred_sign_proxy(canl_ctx ctx, canl_cred signer_cred, canl_cred proxy_cred)
                 (key_size <= DEF_KEY_LEN_LONGER))
             return set_error(cc, CANL_ERR_unknown, CANL_ERROR, "Cannot" 
                     "sign cert. request -the key is too short with "
-                   " respect to cert. lifetime");
+                   "respect to cert. lifetime");
     }
 
     /*TODO flags - limited,version*/
@@ -388,6 +436,7 @@ canl_cred_save_proxyfile(canl_ctx ctx, canl_cred cred, const char *proxy_file)
     int n_certs = 0;
     int i = 0;
     X509 *cert_from_chain = NULL;
+    int cert_in_chain = 0;
 
     if (!ctx)
         return EINVAL;
@@ -424,13 +473,31 @@ canl_cred_save_proxyfile(canl_ctx ctx, canl_cred cred, const char *proxy_file)
 
     ERR_clear_error();
 
-    /*new cert + priv key + chain*/
-    ret = PEM_write_X509(cert_file, crd->c_cert);
-    if (!ret) {
-        ssl_err = ERR_get_error();
-        ret = set_error(cc, ssl_err, SSL_ERROR, "Error while writing"
-               " the certificate to the file");
-        goto end;
+    /*new cert + priv key + chain
+      if the new cert is empty, take it from the chain*/
+    if (crd->c_cert){
+        ret = PEM_write_X509(cert_file, crd->c_cert);
+        if (!ret) {
+            ssl_err = ERR_get_error();
+            if (ssl_err)
+                ret = set_error(cc, ssl_err, SSL_ERROR, "Error while writing"
+                        " the certificate to the file");
+            goto end;
+        }
+    }
+    else if (crd->c_cert_chain){
+        cert_from_chain = sk_X509_value(crd->c_cert_chain, 0);
+        if (cert_from_chain) {
+            ret = PEM_write_X509(cert_file, cert_from_chain);
+            if (!ret) {
+                ssl_err = ERR_get_error();
+                if (ssl_err)
+                    ret = set_error(cc, ssl_err, SSL_ERROR, "Error "
+                            " while writing the certificate to the file");
+                goto end;
+            }
+            cert_in_chain = 1;
+        }
     }
     ret = PEM_write_PrivateKey(cert_file, crd->c_key, NULL, NULL, 0, 0, NULL);
     if (!ret) {
@@ -441,7 +508,7 @@ canl_cred_save_proxyfile(canl_ctx ctx, canl_cred cred, const char *proxy_file)
     }
 
     n_certs = sk_X509_num(crd->c_cert_chain);
-    for (i = 0; i <  n_certs; i++){
+    for (i = cert_in_chain; i <  n_certs; i++){
         cert_from_chain = sk_X509_value(crd->c_cert_chain, i);
         if (cert_from_chain) {
             ret = PEM_write_X509(cert_file, cert_from_chain);
@@ -466,12 +533,12 @@ canl_cred_save_proxyfile(canl_ctx ctx, canl_cred cred, const char *proxy_file)
 end:
     if (fclose(cert_file)){
         ret = errno;
-        update_error(cc, ret, POSIX_ERROR, "cannot close file with certificate");
+        update_error(cc, ret, POSIX_ERROR, "cannot close file"
+                " with certificate");
         return errno;
     }
 
     return ret;
-
 }
 
 canl_err_code CANL_CALLCONV
@@ -491,7 +558,6 @@ canl_cred_save_cert(canl_ctx ctx, canl_cred cred, X509 ** cert)
                 " handler");
  
     if (*cert) {
-        X509_free(*cert);
         *cert = NULL;
     }
 
@@ -625,6 +691,120 @@ canl_cred_load_req(canl_ctx ctx, canl_cred cred_out, const X509_REQ *req_in)
     return 0;
 }
 
+canl_err_code CANL_CALLCONV
+canl_verify_chain(canl_ctx ctx, X509 *ucert, STACK_OF(X509) *cert_chain,
+        char *cadir)
+{
+    int ret = 0;
+    proxy_verify_desc *pvd = NULL; /* verification context */
+
+    pvd = pvd_setup_initializers(cadir);
+    ret = proxy_verify_cert_chain(ucert, cert_chain, pvd);
+    pvd_destroy_initializers(pvd);
+    if (ret)
+        /* This will be ommited when proxy_verify_cert sets errors itself or
+           propagate them out. */
+        return set_error(ctx, CANL_ERR_unknown, CANL_ERROR, "Certificate chain"
+                " validation failed"); // TODO error code check
+    return 0;
+}
+
+canl_err_code CANL_CALLCONV
+canl_verify_chain_wo_ossl(canl_ctx ctx, char *cadir,
+	X509_STORE_CTX *store_ctx)
+{
+    int ret = 0, depth = 0, i = 0;
+    STACK_OF(X509) *certstack;
+    proxy_verify_desc *pvd = NULL; /* verification context */
+    unsigned long ssl_err = 0;
+    canl_error canl_err = 0;
+
+    pvd = pvd_setup_initializers(cadir);
+    X509_STORE_CTX_set_ex_data(store_ctx, PVD_STORE_EX_DATA_IDX, (void *)pvd);
+#ifdef X509_V_FLAG_ALLOW_PROXY_CERTS
+    X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_ALLOW_PROXY_CERTS);
+#endif
+    pvd_destroy_initializers(pvd);
+
+    certstack = X509_STORE_CTX_get_chain(store_ctx);
+    depth = sk_X509_num(certstack);
+    /*TODO maybe free() certstack? is it refferenced? */
+
+    ERR_clear_error();
+    /* Go through the client cert chain and check it */
+    for (i = depth - 1; i >= 0; i--){
+        ret = proxy_verify_callback(1, store_ctx);
+        if (!ret){
+            /* Verification failed */
+            ssl_err = ERR_get_error();
+            canl_err = map_verify_result(ssl_err, store_ctx, NULL);
+            if (canl_err)
+                return set_error (ctx, canl_err, CANL_ERROR,
+                        "Error during SSL handshake");
+            else
+                return set_error(ctx, ssl_err, SSL_ERROR,
+                        "Error during SSL handshake");
+        }
+    }
+
+    return 0;
+}
+
+proxy_verify_desc *pvd_setup_initializers(char *cadir)
+{
+    proxy_verify_ctx_desc *pvxd = NULL;
+    proxy_verify_desc *pvd = NULL;
+    char *ca_cert_dirn = NULL;
+    int err = 0;
+
+    pvd  = (proxy_verify_desc*)     malloc(sizeof(proxy_verify_desc));
+    pvxd = (proxy_verify_ctx_desc *)malloc(sizeof(proxy_verify_ctx_desc));
+    pvd->cert_store = NULL;
+
+
+    if (!pvd || !pvxd) {
+        free(pvd);
+        free(pvxd);
+        return NULL;
+    }
+
+    proxy_verify_ctx_init(pvxd);
+    proxy_verify_init(pvd, pvxd);
+
+    /* If cadir is not specified, do the best as to get the 
+       standard CA certificates directory name */
+    if (!cadir){
+        err = proxy_get_filenames(0, NULL, &ca_cert_dirn, NULL, NULL, NULL);
+        if (!err){
+            pvd->pvxd->certdir = ca_cert_dirn;
+            return pvd;
+        }
+    }
+    else
+        pvd->pvxd->certdir = strdup(cadir);
+    return pvd;
+}
+
+void pvd_destroy_initializers(void *data)
+{
+    proxy_verify_desc *pvd = (proxy_verify_desc *)data;
+
+    if (pvd) {
+        if (pvd->pvxd)
+            proxy_verify_ctx_release(pvd->pvxd);
+
+        free(pvd->pvxd);
+        pvd->pvxd = NULL;
+        proxy_verify_release(pvd);
+
+        /* X509_STORE_CTX_free segfaults if passed a NULL store_ctx */
+        if (pvd->cert_store)
+            X509_STORE_CTX_free(pvd->cert_store);
+        pvd->cert_store = NULL;
+
+        free(pvd);
+    }
+}
 
 #if 0
 canl_err_code CANL_CALLCONV
@@ -634,3 +814,41 @@ canl_req_get_pair(canl_ctx, canl_x509_req, EVP_PKEY **)
 }
 #endif
 
+canl_err_code CANL_CALLCONV
+canl_cred_load_priv_key_pkcs11(canl_ctx ctx, canl_cred cred, const char *label,
+			       canl_password_callback pass_clb, void *arg)
+{
+    int ret;
+    creds *crd = (creds*) cred;
+    unsigned long hSession;
+
+    ret = sc_init(&hSession, NULL, NULL, NULL, CKU_USER, 0);
+    if (ret)
+	return set_error(ctx, EINVAL, POSIX_ERROR, "Failed to open session to smartcard");
+
+    ret = sc_get_priv_key_obj_by_label(hSession, label, &crd->c_key);
+    if (ret)
+	return set_error(ctx, EINVAL, POSIX_ERROR, "Failed to locate private key for '%s' on smartcard",
+			 label);
+
+    return 0;
+}
+
+canl_err_code CANL_CALLCONV
+canl_cred_load_cert_pkcs11(canl_ctx ctx, canl_cred cred, const char *label)
+{
+    int ret;
+    creds *crd = (creds*) cred;
+    unsigned long hSession;
+
+    ret = sc_init(&hSession, NULL, NULL, NULL, CKU_USER, 0);
+    if (ret)
+	return set_error(ctx, EINVAL, POSIX_ERROR, "Failed to open session to smartcard");
+
+    ret = sc_get_cert_obj_by_label(hSession, label, &crd->c_cert);
+    if (ret)
+	return set_error(ctx, EINVAL, POSIX_ERROR, "Failed to locate X.509 certificate for '%s' on smartcard",
+			 label);
+
+    return 0;
+}
